#!/usr/bin/python

"""
Fox Experiment
by Kris Schnee

The "stupid prototype":
There's a grid map. You can move around on it, turn-based.
(Might later want to make this real-time though w/pause button.)
You can press a button to "build" on your current land, which shores it up
and makes more land appear.
A score counter shows the total land size.
Bad guys appear on the fringes and shred the map.
Moving onto bad guys kills 'em.

"""

__AUTHOR__ = "Kris Schnee"
__LICENSE__ = ""

##### IMPORTED MODULES #####

## Built-in
import time
import random

## Third-party
import pygame

## Mine
import art ## Art asset management
import ringtale ## Game state management
import conch ## Sound/Music
import driftwood ## UI

##### GLOBALS #####

WORLD_SIZE_X = 12
WORLD_SIZE_Y = 10

TILE_SIZE = 50

TERRAIN_TYPES = ["plains","mountain","volcanic","forest","wyld forest","lake","sea"]

DIRECTIONS_LIST = [(0,-1),(1,0),(0,1),(-1,0)]

SILLY_LIST = ["Trashed","Toasted","Terminated","Thrashed","TKOed","Terrorized","Told Off","Taunted","Totaled"]


##### CLASSES #####

class Character:
    """A creature such as a monster or bandit."""
    def __init__(self,**options):
        self.name = options.get("name","Person")
        self.coords = options.get("coords",(0,0))
        self.wounds = [0,0,0] ## Light, moderate, critical

        ## Temp
        self.health = 10
        self.health_max = 10

        ## Graphics-related
        ## Uses tiles rather than sprite_sheets.
        self.sprite_sheet = art.tiles[ options.get("sprite_sheet","default") ]

    def Hurt(self,n):
        self.health = max(0,self.health-n)

    def Heal(self,n):
        self.health = min(self.health_max,self.health+n)

    def GetHealth(self):
        """Return a string like "Health: 5/5"."""
        return "Health: " + str(self.health) + "/" + str(self.health_max)




class Land:
    def __init__(self,**options):
        self.coords = options.get("coords",[0,0])
        self.terrain = options.get("terrain","plains")
        self.population = options.get("population",0)
        self.race = options.get("race","human")
        self.ruler = options.get("ruler",None)
        self.wealth = options.get("wealth",0)
        self.lore = options.get("lore",0) ## Bonus research
        self.mana = options.get("mana",0) ## Bonus mana
        self.sanctum = options.get("sanctum",False) ## Is a ruler's HQ?

        self.armies = {}

        ## For demo. Harmony = stability/reality, key feature of the world.
        self.harmony = options.get("harmony",100)

        self.characters = [] ## Characters present here.

    def AddCharacter(self,character):
        if character in self.characters:
            return False
        self.characters.append(character)
        character.coords = self.coords[:]
        return True

    def RemoveCharacter(self,character):
        if character not in self.characters:
            return False
        else:
            self.characters.remove(character)
            character.coords = [None,None]
            return character


class World:
    def __init__(self,**options):
        """The game world, containing Lands."""
        self.lands = []
        self.extant_lands = 0 ## Lands that currently exist (have harmony > 0).
        ## Create blank 2D array.
        for x in range(WORLD_SIZE_X):
            self.lands.append([])
            for y in range(WORLD_SIZE_Y):
                self.lands[-1].append( Land(coords=[x,y]) )

        self.messages = [] ## Redundant?
        self.characters = []

        self.next_demon_spawn_time = time.time()

    def Notify(self,event):
        """Add to a stack of messages (any format OK).

        Must be duplicated here despite appearing in Ringtale.Framework."""
        self.messages.append(event)

    def AddCharacter(self,**options):
        """Make Character, add to record of extant Characters, add to appropriate Land."""
        options.setdefault("coords",[5,5])
        x,y = options["coords"]
        c = Character(**options)
        self.characters.append(c)
        self.lands[x][y].AddCharacter(c)
        return c

    def BuildRandomWorld(self,seed=time.time()):
        random.seed(seed)
        for x in range(WORLD_SIZE_X):
            for y in range(WORLD_SIZE_Y):
                n = random.randint(0,99)
                if n < 25:
                    t = "lake"
                elif n < 50:
                    t = "plains"
                elif n < 75:
                    t = "forest"
                else:
                    t = "mountain"
                self.lands[x][y].terrain = t
        for i in range(3):
            for x in range(1,WORLD_SIZE_X-1):
                for y in range(1,WORLD_SIZE_Y-1):
                    n = random.randint(0,99)
                    if n < 20:
                        continue
                    elif n < 40:
                        t = self.lands[x][y-1].terrain
                    elif n < 60:
                        t = self.lands[x][y+1].terrain
                    elif n < 80:
                        t = self.lands[x-1][y].terrain
                    else:
                        t = self.lands[x+1][y].terrain
                    self.lands[x][y].terrain = t

        ## For demo: Make bits of the world unreal.
        for x in range(WORLD_SIZE_X):
            for y in range(WORLD_SIZE_Y):
                d = max(0,( abs(x-5) + abs(y-5) ))
                self.lands[x][y].harmony = max(0, 100 - (30*d) + random.randint(-40,0))

        ## Build a Sanctum in the middle.
        self.lands[5][5].sanctum = True

    def MoveCharacter(self,character,dx,dy):
        """Tests whether move is allowed. Places character in new Land."""
        x,y = character.coords
        newx, newy = x+dx, y+dy
        if (newx<0) or (newx>=WORLD_SIZE_X) or (newy<0) or (newy>=WORLD_SIZE_Y):
            return False
        target_land = self.lands[newx][newy]
        if target_land.harmony == 0:
            return False ## Can't enter non-existent lands, yet.
        
        self.lands[x][y].RemoveCharacter(character)
        target_land.AddCharacter(character)
        return True

    def CreateDemon(self):
        """Randomly add a bad guy to the map's edges."""
        ## Do it the dumb way first. Pick an empty fringe spot randomly.
        site_ok = False
        tries = 0
        while (not site_ok) and (tries < 50):
            tries += 1
            x = random.randint(0,WORLD_SIZE_X-1)
            y = random.randint(0,WORLD_SIZE_Y-1)
            if (self.lands[x][y].harmony > 0) and (self.lands[x][y].harmony <= 25):
                if self.hero.coords != [x,y]:
                    self.AddCharacter(name="Tengu"+str(random.randint(0,999)),coords=[x,y],sprite_sheet="tengu")
                    self.Notify("demon_spawned")
                    return

    def GetAdjacentCoords(self,coords):
        """Return a list of neighboring coordinates (ortho)."""
        r = []
        x,y = coords
        if x != 0:
            r.append([x-1,y])
        if x < WORLD_SIZE_X-1:
            r.append([x+1,y])
        if y != 0:
            r.append([x,y-1])
        if y < WORLD_SIZE_Y-1:
            r.append([x,y+1])
        return r

    def RunWorld(self):
        """Does logic like creating monsters over time."""

        ## Count the extant Lands.
        self.extant_lands = 0
        for x in range(WORLD_SIZE_X):
            for y in range(WORLD_SIZE_Y):
                if self.lands[x][y].harmony > 0:
                    self.extant_lands += 1

        ## Does everybody know what time it is? Tengu Time!
        ## Yes, this is real-time spawning.
        now = time.time()
        if now >= self.next_demon_spawn_time:
            self.next_demon_spawn_time = now + 1
            if (random.randint(0,99) < 75) and (len(self.characters) <= 5):
                self.CreateDemon()

            ## Demons attack! Also linked to real time.
            for c in self.characters[1:]: ## Ie. all but hero
                x,y = c.coords
                land = self.lands[x][y]

                ## Hero can kill demons!
                if self.hero.coords == [x,y]:
                    self.DestroyCharacter(c)
                    self.Notify("demon_killed")
                    ## Hero might take damage.
                    n = random.randint(0,2)
                    self.hero.Hurt(n)
                    self.Notify("update_health")
                else:
                    land.harmony = max(0,land.harmony-4)
                    if land.harmony == 0:
                        ## Land destroyed!
                        land.sanctum = False ## Uh-oh.
                        neighbors = self.GetAdjacentCoords(c.coords)
                        for n in neighbors:
                            n_land = self.lands[n[0]][n[1]]
                            n_land.harmony = max(0,n_land.harmony-10)

                        ## Try moving to an adjacent Land.
                        moved = False
                        for i in range(5):
                            choice = random.randint(0,3)
                            dx, dy = DIRECTIONS_LIST[choice]
                            moved = self.MoveCharacter(c,dx,dy)
                            if moved:
                                break ## Stop trying.
                        if not moved:
                            ## Demon is destroyed, for now.
                            self.DestroyCharacter(c)
                

    def DestroyCharacter(self,character):
        """Completely remove a Character from the world."""
##        print "Trying to destroy a character: " + character.name + " @ " + str(character.coords)
        x,y = character.coords
        try:
            land = self.lands[x][y]
        except:
            raise Exception("Tried to destory a Character with nonsense coordinates.")

        land.RemoveCharacter(character)
        i = self.characters.index(character)
        self.characters = self.characters[:i] + self.characters[i+1:]
        assert character not in self.characters ## Yup, he's gone.

class Game( ringtale.Framework, World ):
    def __init__(self,**options):
        ringtale.Framework.__init__(self,**options)
        World.__init__(self,**options)
        

        pygame.display.set_caption("Prototype Kitsune Game")
        pygame.display.set_icon(art.posters["acorn"])

        self.screen = pygame.display.set_mode((800,500))

        self.SetMusicHandler(conch)

        ## The world coordinates which are drawn as pixel (0,0):
        self.camera_x = 0
        self.camera_y = 0

        self.BuildRandomWorld()

        ## Create PC.
        self.hero = self.AddCharacter(name="Fox",coords=[5,5],sprite_sheet="fox")
        self.hero_demons_killed = 0

        ## Starting game state.
        self.PushState("World")

    def GetScreenCoords(self,world_coords):
        return (world_coords[0] - self.camera_x,
                world_coords[1] - self.camera_y)

    def DrawLandHighlight(self,coords):
        """Draw a scary red border around a land."""
        sx, sy = coords[0]*TILE_SIZE, coords[1]*TILE_SIZE
        pygame.draw.rect(self.screen,(255,0,0),(sx,sy,TILE_SIZE,TILE_SIZE),1)

    def GetClickedCoords(self,coords):
        """Find tile coordinates clicked, or return None."""
        ## NOTE: Currently assumes no scrolling.
        x, y = coords
        if (x<0) or (y<0) or (x>600) or (y>500):
            return None
        return (int(x/TILE_SIZE),int(y/TILE_SIZE))



    def HandleEvents(self,events=[]):
        for event in events:
            ## First, give the widgets a chance to react.
            handled = False
##            for widget in self.interface:
##                handled = widget.HandleEvent(event)
##                if handled:
##                    break

            ## If they don't, handle the event here.
            if not handled:
                if event.type == pygame.locals.QUIT:
                    self.PopState()

                elif event.type == pygame.locals.KEYDOWN:
                    if event.key == pygame.locals.K_ESCAPE:
                        self.JumpState("GameOver")
                        return ## We're done.

                    elif event.key == pygame.locals.K_LEFT:
                        self.Notify("move_w")
                    elif event.key == pygame.locals.K_RIGHT:
                        self.Notify("move_e")
                    elif event.key == pygame.locals.K_UP:
                        self.Notify("move_n")
                    elif event.key == pygame.locals.K_DOWN:
                        self.Notify("move_s")
                    elif event.key == pygame.locals.K_SPACE:
                        self.Notify("build")
                        
                elif event.type == pygame.locals.MOUSEBUTTONDOWN:
                    ## Find where in the world was clicked.
                    world_coords = self.GetClickedCoords(event.pos)
                    if world_coords:
                        print "Clicked tile: " + str(world_coords)
                        self.selected_land = world_coords
                        self.Notify("land_was_selected")

    def WorldSetup(self):
        self.messages = []
        self.interface = []
        self.interface.append( driftwood.TextBox(owner=self,coords=(600,0,200,50),name="lWorldSize",text="World Size: 0") )
        self.interface.append( driftwood.TextBox(owner=self,coords=(600,50,200,50),name="lDemonsKilled",text="Tengu Trashed: 0") )
        self.interface.append( driftwood.TextBox(owner=self,coords=(600,100,200,50),name="lDemonsKilled",text=self.hero.GetHealth()) )
        self.interface_index = {}
        self.interface_index["lWorldSize"] = self.interface[0]
        self.interface_index["lDemonsKilled"] = self.interface[1]
        self.interface_index["lHealth"] = self.interface[2]

    def WorldDraw(self):

        fog = art.tiles["fog"]
        
        for x in range(WORLD_SIZE_X):
            for y in range(WORLD_SIZE_Y):
                land = self.lands[x][y]
                t = art.tiles[ land.terrain ]
                draw_x = x*TILE_SIZE
                draw_y = y*TILE_SIZE

                ## Draw darkness over the land! Mwahaha.
                fog_opacity = (100-land.harmony) * 2.55
                fog.set_alpha(fog_opacity)
                
                self.screen.blit(t,(draw_x,draw_y))
                if land.sanctum:
                    t = art.tiles["tower"]
                    self.screen.blit(t,(draw_x,draw_y))
                elif land.population:
                    t = art.tiles["town"]
                    self.screen.blit(t,(draw_x,draw_y))

                self.screen.blit(fog,(draw_x,draw_y))
                if (land.harmony > 0) and (land.harmony <= 25):
                    ## Draw menacing border.
                    self.DrawLandHighlight((x,y))

        for c in self.characters:
            x, y = c.coords
            try:
                assert x != None
            except:
                raise Exception(("Coords error",c,c.name,c.coords))
            t = c.sprite_sheet
            draw_x = x*TILE_SIZE
            draw_y = y*TILE_SIZE
            self.screen.blit(t,(draw_x,draw_y))

        self.interface_index["lWorldSize"].SetText("World Size: "+str(self.extant_lands))

##        if self.selected_land:
##            self.DrawLandHighlight(self.selected_land)

        for widget in self.interface:
            widget.Draw()
      
    def WorldLogic(self,events):

        ## Events in the game world happen before input is handled.
        self.RunWorld()

        self.HandleEvents(events)
        for message in self.messages:
            if message == "land_was_selected":
                pass
##                self.interface_index["lLandInfo"].SetText( self.WriteLandInfo(self.selected_land) )

            elif message == "move_w":
                self.MoveCharacter(self.hero,-1,0)
            elif message == "move_e":
                self.MoveCharacter(self.hero,1,0)
            elif message == "move_n":
                self.MoveCharacter(self.hero,0,-1)
            elif message == "move_s":
                self.MoveCharacter(self.hero,0,1)

            elif message == "build":
                x,y = self.hero.coords
                land = self.lands[x][y]
                land.harmony += 10
                if land.harmony >= 100:
                    land.harmony = 100
                    neighbors = self.GetAdjacentCoords(self.hero.coords)
                    for n in neighbors:
                        n_land = self.lands[n[0]][n[1]]
                        if n_land.harmony < 25:
                            n_land.harmony = 25
##                    conch.PlaySound("yay")
                if land.sanctum:
                    self.hero.Heal(99)
                    self.Notify("update_health")

            elif message == "demon_spawned":
                pass
            elif message == "demon_killed":
                self.hero_demons_killed += 1
                self.interface_index["lDemonsKilled"].SetText( "Tengu " + SILLY_LIST[random.randint(0,8)] + ": " + str(self.hero_demons_killed ))
##                conch.PlaySound("demon_killed")
            elif message == "update_health":
                self.interface_index["lHealth"].SetText(self.hero.GetHealth())
                if self.hero.health == 0:
                    self.JumpState("GameOver")
                    return

    def GameOverSetup(self):
        self.messages = []
        self.interface = []
        self.interface.append( driftwood.TextBox(owner=self,coords=(100,200,300,100),name="lGameOver",text="*** Game Over! ***\nHit Escape to quit.\nThanks for trying this demo!") )

    def GameOverLogic(self,events):
        for event in events:
            if event.type == pygame.locals.KEYDOWN:
                if event.key == pygame.locals.K_ESCAPE:
                    self.PopState()

    def GameOverDraw(self):
        for widget in self.interface:
            widget.Draw()


##### FUNCTIONS #####


##### AUTORUN #####
if __name__ == "__main__":
    g = Game()
    g.MainLoop()
